#!/usr/bin/wish

# Copyright (c) 2002-2014, Botik Technologies, Ltd.
# See license.txt for more information.

#                      -------- rawTraceroute  ,  
#
#    Windows,  .       .
#
# rawTraceroute  tracerout  host.      n.
# 
#    ping  IP- $ping(trrt_ip,$k)    ping(trrt_h,$k) 
#   ($k=0..n-1).  :    $ping(trrt_ip,N)=n.
#
#     ,  ping(trrt_h,$k)=$ping(trrt_ip,$k).
#
#    n=0.  
#
# :      $ping(raw_trrt,$k),  $k=0...   .
# :    : $ping(raw_trrt,N)=$k+1
#
proc rawTraceroute {host} {
   global ping

   if {[windowsPlatform]} {
        set traceregexpr {^(.*)\s+(([-a-zA-Z0-9_.][-a-zA-Z0-9_.]*)\s+\[(\d+\.\d+\.\d+\.\d+)\]|(\d+\.\d+\.\d+\.\d+))(.*)$}
   #              head^^^^    ^^^^^^^^^host-name^^^^^^^^^^^^^   ^^^^^^ip-address^^^^^  ^^^ip-address^^^ ^^^^tail 
   #                         ^^^^hiORi = host-name+ip-address OR ip-address^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
   } else {
        set traceregexpr {^(.*)\s+(([-a-zA-Z0-9_.][-a-zA-Z0-9_.]*|\d+\.\d+\.\d+\.\d+)\s+\((\d+\.\d+\.\d+\.\d+)\))(.*)$}
           #              head^^^^     ^^^^^^^^host-name^^^^^^^^^^^^ ^^^^ip-address^^^^       ^^^^ip-address^^^^ ^^^^tail 
           #                         ^^^^hORii = host-name OR ip-address + ip-address^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
   }

   set tres "n/a"
   if {$host eq [info hostname]} {
      set ipv6check 1
   } else {
      set ipv6check 0
   }
   if {![info exists ping(isModem)]} {set ping(isModem) 0}
   if {![info exists ping(raw_trrt,N)]} {set ping(raw_trrt,N) 0}
   set n [lindex { "" "-w 3000" } $ping(isModem)]
   if {[windowsPlatform]} {
       if {[WinNT6]} {
           set tracecmd "tracert -4 $n $host"
       } else {
           set tracecmd "tracert $n $host"
       }
   } else {
       set tracecmd "traceroute $n $host"
   }
   if {[btools_exe]} {
       set tres [tracert $host]
   } else {
        catch {
            eval "exec $tracecmd"
        } tres
   }
   if {[btools_exe]} {
        set tres [encoding convertfrom cp866 $tres]
   } else {
        set tres [encoding convertfrom cp866 [encoding convertto cp1251 $tres]]
   }
   regsub -all "\n\n" $tres "\n" traceres
   set n $ping(raw_trrt,N) 
   set ping(raw_trrt,$n) "$tracecmd\n$traceres"
   set ping(raw_trrt,N)  [expr $n + 1]
   set ping(trrt_ip,N) 0
   set i 0
   while { [regexp $traceregexpr $traceres skip head hiORi h ip justip tail] } {
       #debug "hiORi=$hiORi\nh=$h ip=$ip\njustip=$justip\nhead=$head\ntail=$tail\nskip=$skip"
       if { [string length $h] == 0 } {
           set h  $justip
           set ip $justip
       } else {
           if {$ipv6check && [string match *:* $ip]} {
               set ping(ipv6enabled) 1
           }
       }
       if { [chkIP $ip] } {
            set hip($i) $ip 
            set hname($i)  $h 
            incr i
       }
       set traceres "$head $tail"
   }
   set n $i
   
   #    ping(trrt_ip,k)   ping(trrt_h,k)   
   #
   # i = (n-2) (n-3) ... 0     (n-1)      hip($i),           hname($i)
   # k =   0     1     (n-2)   (n-1)      ping(trrt_ip,$k),  ping(trrt_h,$k)
   set k 0
   for {set i [expr $n-2]} {$i >= 0} {set i [expr $i-1]} {
      set k [setPingTrrt $k $hname($i) $hip($i)]
   }
   set i [expr $n-1]
   if {$i >= 0} {
      set k [setPingTrrt $k $hname($i) $hip($i)]
   }
   set ping(trrt_ip,N) $k
   return $k
}
proc setPingTrrt { k hn ip } {
   global ping
   for {set i 0} {$i < $k} {incr i} {
      if { $ping(trrt_ip,$i) eq $ip } {
         if { [llength [split $ping(trrt_h,$i) . ] ] <= [llength [split $hn . ] ] } {
            set ping(trrt_h,$i) $hn
         }
         return $k
      }
   }
   set ping(trrt_ip,$k) $ip
   set ping(trrt_h,$k)  $hn
   return [expr $k + 1]
}

proc rawPing {host {size 32} {timeout 1000} } {
#   { rec (0:1) ms $msec lastPingCmd $pingcmd lastPingRes $pres}
#
#           array set xxx [rawPing $host $size $timeout]
#
#  

   if {[windowsPlatform]} {
        set pingregexpr {TTL=.*Ping.*=\s*\d+.*=\s*(\d+).*=\s*\d+.*=\s*\d+.*=\s*\d+.*=\s*(\d+)}
#                               Send^^^    Rcv^^^   Lost^^^   Min^^^   Max^^^    Avg^^^
   } else {
        set pingregexpr {ms.*\d+.*,\s*(\d+).*,\s*\d+.*=\s*\d+\.\d+/(\d+)\.\d+/\d+.*}
#                      Send^^^    Rcv^^^    Lost^^^   Min^^^^^^^Avg^^^^^^^Max^^^    
   }

   set rec 0
   set ms  0
   set pingres 0
   if {[windowsPlatform]} {
       if {[WinNT6]} {
           set pingcmd "ping -4 -n 1 -l $size -w $timeout $host"
       } else {
           set pingcmd "ping -n 1 -l $size -w $timeout $host"
       }
   } else {
       set timeo [expr $timeout/1000.0]
       if {[macPlatform]} {
           set timeo [expr ceil($timeo)]
       }
       set pingcmd "ping -c 1 -s $size -i $timeo $host"
   }
   if {![info exists ::bt_exe]} {
        catch "exec $pingcmd" pingres
   } else {
        set pingres [ping -n 1 -l $size -w $timeout $host]
   }

   regexp $pingregexpr $pingres skip rec ms
   set rec [limits $rec 0   1]
   set ms  [limits $ms  0 $ms]

   if {[info exists ::bt_exe]} {
       set pingres [encoding convertfrom cp866 $pingres]
   } else {
       set pingres [encoding convertfrom cp866 [encoding convertto cp1251 $pingres]]
   }
   regsub -all "\n\n" $pingres "\n" pres
   return "rec $rec ms $ms lastPingCmd {$pingcmd} lastPingRes {$pres}"
}
#                      --------    
proc pswdToKey { pswd } {
   return [string tolower [md5::md5 -hex -- [encoding convertto koi8-r $pswd]]]
}
#                      --------   ,  IP-
proc chkIP { ip } {
   set x {^([1-2][0-9][0-9]|[1-9][0-9]|[0-9]).([1-2][0-9][0-9]|[1-9][0-9]|[0-9]).([1-2][0-9][0-9]|[1-9][0-9]|[0-9]).([1-2][0-9][0-9]|[1-9][0-9]|[0-9])$}
   set a -1
   set b -1
   set c -1
   set d -1
   return [expr [regexp $x $ip skip a b c d]\
                &&  (-1 < $a) && ($a < 256) \
                &&  (-1 < $b) && ($b < 256) \
                &&  (-1 < $c) && ($c < 256) \
                &&  (-1 < $d) && ($d < 256) \
          ]
}
#                      --------  $text     ":", 
#                                (regsub -all) $bls  
proc scolSplit { text {bls ""} } {
   set t $text

# WhoAmI  Time   "// ::",    ,
#    .      
# "  ( ):...   ( )".
#   ,      (scolSplit)
#  .

    set tokens [split $t \n]
    set list   [list]
    # 1.   :
    foreach el $tokens {
        if {[llength $el] > 0} {
            lappend list $el
        }
    }
    set list2 [list]
    foreach el $list {
        set tmp $el
        # 2.  ":"   :
        if {[string index $el 0] eq ":"} {
            set tmp [string range $tmp 1 end]
        }
        # 3.  ":"   :
        if {[string index $tmp end] eq ":"} {
            set tmp [string range $tmp 0 end-1]
        }
        # 4.    Time,   
        #      (  ):
        set spl [split $tmp :]
        if {[llength $spl] > 2} {
            set par [lindex $spl 0]
            set len [string length $par]
            if {$par eq "Time"} {
                set val [string range $tmp [expr $len+1] end]
                lappend list2 $par $val
            } else {
                foreach elem $spl {
                    lappend list2 $elem
                }
            }
        } else {
            lappend list2 [lindex $spl 0] [lindex $spl 1]
        }
    }
    return $list2    

   regsub -all " |\t|\n" $t {} x

   #  $bls  " ":
   if { [string length $bls] > 0 } {
      regsub -all $bls $x " " x
   }

   set tokens [split $x :]

   #    :
   set i [lsearch -exact $tokens {}]
   while { ([llength $tokens] > 0) && ($i > -1) } {
      set tokens [lreplace $tokens $i $i]
      set i [lsearch -exact $tokens {}]
   }
   return $tokens
}
#                      --------  ,    ,   
#                                :  bad_file_or_token (0 | 1)  bad_file_or_token_msg
#
proc readTokensFromFile {file} {
   global bad_file_or_token
   global bad_file_or_token_msg

   set bad_file_or_token     0
   set bad_file_or_token_msg ""

   set ok 0
   catch {
      set f [open $file r]
      set ping(file) {}
      set t ""     
      while {![eof $f]} {
          set t "$t[read $f 10000]"
      }
      close $f
      set ok 1
   } rc
   if { ! $ok } {
      checkToken 0 "Error: [mkErrorCode $rc]"
      return {}
   }
   return [scolSplit $t]
}
#                      --------    ,  ( bad_file_or_token  bad_file_or_token_msg) 
#                                 ,    
proc checkToken { condition msg } {
   global bad_file_or_token
   global bad_file_or_token_msg

   if { ($bad_file_or_token==1) || ($condition==1) } { 
      return 
   }
   set bad_file_or_token 1
   set bad_file_or_token_msg $msg
   return
}
#                      --------   x  xmin ... xmax
proc limits { x xmin xmax } {
   return [expr ( $x < $xmin )  ? $xmin : ( $x > $xmax ? $xmax : $x ) ]
}
#                      -------- min
proc min { a b } {
   return [expr ($a < $b)  ? $a  : $b ]
}
#                      -------- max
proc max { a b } {
   return [expr ($a > $b)  ? $a  : $b ]
}
proc alertQuestion {msg {def "yes"}} {
    return [tk_messageBox -icon question -type yesno -default $def -title "" -parent . -message $msg]
}
proc alert {msg} {
    message info $msg
}
#                      --------  , $tag == error question warning info 
rename message message_; # 'message_' is used in 'proc balloon_show'
proc message { tag str } {
   set ttl(info)     ""
   set ttl(warning)  ""
   set ttl(error)    ""
   set ttl(question) ""
   if {![lcontain {error info question warning} $tag]} {
      set title ""
   } else {
      set title $ttl($tag)
   }
   tk_messageBox -icon $tag -type ok -message $str -title $title
}
#                      --------   , $tag == error question warning info 
proc showMessage { tag txt } {
   message $tag [join [splitWrapLines 80 $txt] "\n"]
   return
}
#                      --------      
proc strTime {} {
    return [clockFormat "%d/%m/%Y %H:%M:%S"]
}
proc strTime2 {} {
    return [clockFormat %Y%m%d%H%M%S]
}
proc strTime3 {time} {
    set res ""
    catch {
        set res [clockFormat "%H:%M:%S" $time]
    }
    return $res
}
#                      --------  ( \n) -->  
proc wrapLines {len s} {
   set s [string trimright $s]
   set t ""
   set i1 [expr $len - 1]
   while { [string length $s] > $len } {
      set s1 [string range $s 0 $i1]
      set s3 [string range $s $len end]
      set j  [string last " " $s1]
      if { $j == -1 } {
         set t "$t$s1\n"
         set s $s3
      } else {
         set s11 [string range $s1 0 [expr $j - 1]]
         set s2  [string range $s1 [expr $j + 1] end]
         set t "$t$s11\n"
         set s "$s2$s3"
      }
   }
   set t "$t$s\n"
   return $t
}
#                      --------   \n   " // " -->   
proc splitWrapLines { len txt } {
   regsub -all " // " $txt  "\n" txt1
   regsub -all "\n\n" $txt1 "\n" text

   #  :
   set pars [split $text "\n"]

   #    :
   set i [lsearch -regexp $pars {^\s*$}]
   while { ([llength $pars] > 0) && ($i > -1) } {
      set pars [lreplace $pars $i $i]
      set i [lsearch -exact $pars {}]
   }

   #         .
   set f 1
   set lines ""
   foreach par $pars {
      set p_lines [wrapLines $len [string trimright $par ] ]
      if { $f } {
         set lines $p_lines
      } else {
         set lines "$lines\n$p_lines"
      }
      set f 0
   }
   set lines [split $lines "\n"]
   set n [llength $lines] 
   if {[string is integer -strict $n] && $n > 1} {
      set lines [lrange $lines 0 [expr $n - 2]]
   }
   return $lines
}
#                      --------     len;   "..." 
proc leftDots {len str} {
   set l [string length $str]
   if {$l <= $len} {
      return $str; 
   }
   return "...[string range $str [expr $l+3-$len] end]"
}
#                      --------     len;   "..." 
proc rightDots {len str} {
   set l [string length $str]
   if {$l <= $len} {
      return $str;
   }
   return "[string range $str 0 [expr $len-4]]..."
}

# ================= Tk : ===================================================

proc setImgButtons { w padx pady side a bs } {
   foreach {n img cmd} $bs {
      button $w.b$n -image $img -command $cmd \
             -background gray84 -activebackground LightSteelBlue -highlightbackground LightSteelBlue
      pack $w.b$n -expand 1 -padx $padx -pady $pady -side $side -anchor $a
      setBalloon $w.b$n $img
   }
   return
}
proc setBalloon {b img} {
    if {[info exists ::botikkey(showballoons)] && !$::botikkey(showballoons)} {return}
    if {[info exists ::hints($img)]} {balloon $b $::hints($img)}
}
proc balloon {w helptext} {
    bind $w <Any-Enter> "after 500 [list balloon_show %W [list $helptext]]"
    bind $w <Any-Leave> "destroy %W.balloon"
}
proc balloon_show {w arg} {
    if {[eval winfo containing [winfo pointerxy .]]!=$w} {return}
    set top $w.balloon
    catch {destroy $top}
    toplevel $top -bd 1 -bg black
    wm overrideredirect $top 1
    if {$::tcl_platform(platform) == "macintosh"} {
        unsupported1 style $top floating sideTitlebar
    }
    pack [message_ $top.txt -aspect 700 -bg lightyellow -text $arg]
    set wmx [winfo rootx $w]
    set wmy [expr [winfo rooty $w]+[winfo height $w]]
    wm geometry $top [winfo reqwidth $top.txt]x[winfo reqheight $top.txt]+$wmx+$wmy
    raise $top
}
proc findExecutable {progname varname} {
    upvar 1 $varname result
    set progs [auto_execok $progname]
    if {[llength $progs]} {
        set result [lindex $progs 0]
    }
    return [llength $progs]
}
proc findBrowser {} {
    if {[info exists ::env(BROWSER)]} {
        return $::env(BROWSER)
    }
    # a lot of GUI browsers exist for Linux/UNIX
    set browsers {firefox mozilla-firefox epiphany epiphany-browser seamonkey iceweasel \
                icecat iceape chimera2 galeon mozilla konqueror hotjava \
                netscape opera dillo netsurf flock-browser conkeror \
                kazehakase midori strata skipstone swiftfox amaya google-chrome \
                x-www-browser safari xdg-open}
    set tmp {}
    foreach bro $browsers {
        if {[findExecutable $bro tmp]} {
            set ::env(BROWSER) $tmp
            return $tmp
        }
    }
    return
}
#                      -------- showURL  ,  
#
#    Windows,  .   $env(COMSPEC) --- .
#   "tmp.url" (.)  ( )    
#   
# 
#      URL
#

proc showURL {url} {
 if {[unixPlatform]} {
    findBrowser
    if {[info exists ::env(BROWSER)] && [catch {exec $::env(BROWSER) $url} err]} {
        message error "   URL (by $::env(BROWSER)).\n$url\n$err"
        asksendmail "   URL (by $::env(BROWSER)).\n$url\n$err"
    }
 } elseif {[windowsPlatform]} {
    # Win95 and earlier
    if {$::tcl_platform(osVersion) <= 4.0} {
        showURL2 $url
        return
    }
    # Win98
    if {$::tcl_platform(osVersion) == 4.10} {
        regsub -all "/" $url "\\\\" url
    }
    regsub -all "&" $url "^&" url
    if {$::tcl_platform(osVersion) < 5.0} {
        # Win 98/ME
        if {[catch {
            eval exec [auto_execok start] [list $url]
        } rc]} {
            message error "   URL.\n$url\n\nError: [mkErrorCode $rc]"
            asksendmail "   URL.\n$url\n\nError: [mkErrorCode $rc]"
        }
    } else {
        # Windows NT (e.g. 2K+C/XP/Vista/7/)
        if {[string match *.htm $url]} {
            # rundll32       .htm,     .ht%6D
            set url [regsub -nocase {\.htm} $url {\.ht%6D}]
        }
        set path [auto_execok rundll32]
        if {![lempty $path] && [catch {
            exec rundll32 url.dll,FileProtocolHandler $url
        } rc]} {
            message error "   URL.\n$url\n$path\nError: [mkErrorCode $rc]"
            asksendmail "   URL.\n$url\n$path\nError: [mkErrorCode $rc]"
            showURL2 $url
        }
    }
 } elseif {[macPlatform]} {
    exec open $url; # this just opens default web browser on Mac OS
 }
 return
}
#  (  )      Windows
#     
proc showURL2 {url} {
    set x $::env(COMSPEC)
    if {$x eq ""} {
        set msg "        COMSPEC."
        message error $msg
        asksendmail $msg
        return 1
    }
    if {[catch {
        set tmpfile [tempFile tmp.url]
        set f [open $tmpfile w]
        puts $f "\[InternetShortcut]\nURL=$url"
        close $f
    } rc]} {
        set msg "   URL.\n$url\n\nError: [mkErrorCode $rc]"
        message error $msg
        asksendmail $msg
        return 1
    }
    catch {exec $x /c start $tmpfile}
    #catch {file delete -force $tmpfile}
    return 0
}
proc tempFile { fname } {
   global env
   set d {}
   catch { set d $env(TEMP) }
   if {[string length $d] == 0} { catch { set d $env(TMP) } }
   set tmpfile $fname
   if {[string length $d] != 0} {
      set tmpfile [file join $d $fname]
   }
   return $tmpfile
}
proc appFile {fname {curdir 0}} {
   if {[windowsPlatform]} {
      if {$curdir} {return $fname}
      if {[info exists ::env(APPDATA)]} {
         set dir $::env(APPDATA)
      } else {set dir {}}
      set folder "BotikTools"
   } else {
      set dir    $::env(HOME)
      set folder ".BotikTools"
      if {$curdir} {return [file join $folder $fname]}
   }
   set appFile $fname
   if {![lempty $dir]} {
      catch {
         set appPath [file join $dir $folder]
         file mkdir $appPath
         set appFile [file join $appPath $fname]
      }
   }
   return $appFile
}
#   
proc connTest {{hostname www}} {
   set n    5
   set lost 0
   set rec  0
   for {set i 0} {$i < $n} {incr i} {
      array set ping [rawPing $hostname.$::net(domain)]; # ping'ing www.botik.ru $n times
      if {$ping(rec)} {
         incr rec
      } else {
         incr lost
      }
   }
   if {$lost == $n} {return 0}
   return 1
}

# =================     HTTP: =========================

package require http 2.4 
package require tls  1.4 
http::register https 443 ::tls::socket

proc btools_exe {} {
    return [expr {[info exists ::bt_exe] && $::bt_exe}]
}
proc readURL {url timeout} {
    global ping
    catch {
        set ping(current_request) [http::geturl $url -command chkReadUrl -timeout $timeout]
    }
    return
}
proc chkReadUrl {token} {
   upvar #0 $token state

   set err  ok
   set data ""
   set url  ""
   catch {set err $state(error)}
   catch { 
      set rc "<$err><$state(status)><$state(http)>"
      if {$rc eq "<ok><ok><HTTP/1.1 200 OK>"} {
         set data $state(body)
         set url $state(url)
         analyzeDataFromUrl $url $data
      } elseif {$rc eq "<ok><ok><HTTP/1.1 302 Moved>"} {
         array set meta $state(meta)
         after 500 readURL $meta(location) 5000
      }
   }
   catch {http::cleanup $token}
   return
}
proc windowsPlatform {} {
   return [expr {$::tcl_platform(platform) eq "windows"}]
}
proc unixPlatform {} {
   return [expr {$::tcl_platform(platform) eq "unix" && $::tcl_platform(os) ne "Darwin"}]
}
proc macPlatform {} {
   return [expr {$::tcl_platform(os) eq "Darwin"}]
}
proc sendRouteInfo {} {
    set route ""
    set arp ""
    if {[info exists ::routesent]} {
        alert "      ."
        return
    }
    catch {
    if {[windowsPlatform]} {
        set route [exec route print]
    } else {
        set route [exec route]
    }
    }
    catch {
        set arp [exec arp -a]
    }
    if {[llength $route] || [llength $arp]} {
        set ans [alertQuestion "  \"$::net(name)\"         .   ?"]
        if {$ans eq "no"} {return}
        set msg "Routing table:\n$route\n\nARP info:\n$arp"
        sendmail $msg
        set ::routesent 1
    }
}
# rawGateway  IP ,  -1
proc rawGateway { } {
   set ip -1
   set cmdres ""
   if {[windowsPlatform]} {
      # Windows specific
      set cmd "route print"
      set re {^.*\n\s+0\.0\.0\.0\s+\s+0\.0\.0\.0\s+(\d+\.\d+\.\d+\.\d+)\s.*$}
   } elseif {[unixPlatform]} {
      # Linux specific
      set cmd "netstat -rn" 
      set re {^.*\n0\.0\.0\.0\s+(\d+.\d+.\d+.\d+)\s.*$}
   } elseif {[macPlatform]} {
      # Mac OS X specific
      set cmd "netstat -rn" 
      set re {^.*\ndefault\s+(\d+.\d+.\d+.\d+)\s.*$}
   } else {
      #writeLogChangeStatus 1 RED "[strTime] : rawGateway:   windows   unix"
      return -1
   }
   catch "exec $cmd" cmdres
   regexp $re $cmdres skip ip
   if [chkIP $ip] {
      return $ip
   } else {
      return -1
   }
}

proc getWhoAmIInfo {} {
    whoAmI
    return [formatWhoAmIData]
}

#     WhoAmI
proc whoAmI {} {
    global whoami botikkey
    array unset whoami
    set botikkey(time) ""
    if {[connTest]} {
        catch {http::geturl $botikkey(url_whoami) -handler whoAmIHandler -timeout 10000}
    }
}

#     WhoAmI
proc whoAmIHandler {socket token} {
    global whoami botikkey
    gets $socket line
    set spl   [split $line ":"]
    set kword [lindex $spl 1]
    set value [string range $line [expr [string length $kword] + 2] end]
    set whoami($kword) $value
    if {$kword eq "Time"} {
        set botikkey(time) [ftime]
    }
    return
}
#   
proc ftime {{t {}}} {
    return [clockFormat "%Y/%m/%d %X" $t]
}
#  WhoAmI-
proc formatWhoAmIData {} {
    global whoami botikkey
    set myinfo ""
    if {[array exists whoami]} {
        set all [array get whoami]
        foreach {elt val} $all {
            if {$elt eq ""} {continue}
            set myinfo "$myinfo\n:$elt:$val"
            if {$elt eq "Time"} {set myinfo "$myinfo\n:Client time:$botikkey(time)"}
        }
    }
    return $myinfo
}
# Determines whether we have Windows NT kernel v.6.0 or later (Windows Vista, Windows 7, etc.)
proc WinNT6 {} {
    return [expr {[windowsPlatform] && $::tcl_platform(osVersion) >= 6.0}]
}
# Determines whether we have Windows 98 kernel v.4.1 or earlier
proc Win98 {} {
    return [expr {[windowsPlatform] && $::tcl_platform(osVersion) <= 4.10}]
}
proc tokoi {str} {
    return [encoding convertto koi8-r $str]
}
proc towin {str} {
    return [encoding convertto cp1251 $str]
}
proc dbg {str} {
    set f [open debug.txt a]
    puts $f $str
    close $f
}
proc desktop_dir {} {
    global ddir
    if {![info exists ddir]} {set ddir ""}
    if {$ddir ne ""} {return $ddir}
    set dir ~
    if {[windowsPlatform]} {
        catch {
            package require twapi
            set dir [twapi::get_shell_folder desktopdirectory]
        }
    }
    set ddir $dir
    return $dir
}
if {[windowsPlatform]} {
    package require tile
    namespace import -force ttk::*
}
if {[unixPlatform]} {
    catch {source bgerror.tcl}
}
