#!/usr/bin/wish

# Copyright (c) 2002-2014, Botik Technologies, Ltd.
# See license.txt for more information.

package require Tcl  8.2
package require md5  2.0.0

# ===========================================================================================
proc mkErrorCode { rc {delim "\n"}} {
   global errorCode
   set x $errorCode
   set t [lindex $x 0]
   switch $t {
      ARITH       -
      CHILDKILLED -
      CHILDSUSP   - 
      POSIX       { return "[lindex $x end]$delim$rc" }
      default     { return $rc }
   }
}

proc loadModule { fname } {
   global myModules
   if [info exists myModules($fname)] {
      tk_messageBox -icon warning -type ok -message " \"$fname\"  "
      return
   }
   set ok 0
   catch {
      source $fname
      set ok 1
      set myModules($fname) 1
   } rc
   if { ! $ok } {
      tk_messageBox -icon error -type ok -message \
            "     \"$fname\"\nError: [mkErrorCode $rc]\n\n\
                .\n\
               ."
      exit
   }
   return
}

proc loadImage { name fname } {
   global myImages
   if [info exists myImages($name)] {
      return
   }
   set ok 0
   catch {
      set myImages($name) [image create photo -file $fname]
      set ok 1
   } rc
   if { ! $ok } {
      tk_messageBox -icon error -type ok -message \
            "      : \"$fname\"\nError: [mkErrorCode $rc]\n\n\
                .\n\
               ."
      exit
   }
   return
}

proc debug { str } {
   message info "DEBUG $str"
}

# ===========================================================================================
loadModule bkInit.tcl
loadModule botikUts.tcl
loadModule net.tcl
loadModule report.tcl

loadImage  fnameNormal       img/pswd_write.gif
loadImage  noNormal          img/no.gif

# =================   =============================================

proc mkKey { } {
   after 10 mkKey1
   return 1
}

proc mkKey1 { } {
   global pswd key
   set key [pswdToKey  $pswd]
   return
}

proc writeKey { } {
   global pswd key botikkey

   set key  [pswdToKey  $pswd]

   set types {
       {{All Files}        *             }
       {{Text Files}       {.txt}        }
   }

   set idir [file dirname $botikkey(dflt_pswd_file)]
   set writeFileName [file tail $botikkey(dflt_pswd_file)]
   set filename [tk_getSaveFile -filetypes $types -defaultextension * -parent .mw \
                                -title "    ?" \
                                -initialfile $writeFileName -initialdir $idir]
   if {![llength $filename]} {
      message info "      ."
      return
   }

   # $filename --  
   set goon "retry"
   set ok 0
   while {$goon eq "retry"} {
      set ok 0
      catch {
         #     $filename
           set out [open $filename w]
         puts $out $key
         close $out
         set ok 1
      } rc
      if { !$ok } {
         catch { close $out }
         set msg "      \n\"$filename\"\n\n[mkErrorCode $rc]"
         set goon [ tk_messageBox -icon warning -type retrycancel -default retry \
                                  -title "     " -parent .mw \
                                  -message $msg ]
         asksendmail $msg
      } else {
         set goon cancel
      }
   }
   if { !$ok } {
      message info "      ."
      return
   }
   message info "     :\n$filename"
   return
}
proc genPw {{pwlen 9}} {
#  - .
#      ISAAC.
#           .
    package require Random 1.2.0
    if {![info exists ::track] || $::track eq {}} {return}
    if {![info exists ::seed]} {
        append ::seed [clock clicks] [winfo id .] $::track
        random::isaac seed $::seed
        bind . <Motion> ""
        set ::seed {}
    }

    set chars [split {~!@#$%^&*()+aAbBcCdDeEfFgGhHjJkKLmMnNpPqQrRsStTwWxXyYzZ23456789} ""]; #    
    set len   [llength $chars]
    set pw    {}

    while {[string length $pw] < $pwlen} {
        set rand [expr abs([random::isaac integer])]
        while {$rand > 0 && $pwlen && [string length $pw] < $pwlen} {
            set rest [expr {$rand % $len}]
            set rand [expr {$rand / $len}]
            append pw [lindex $chars $rest]
        }
    }
    set re1 {[a-z]+}
    set re2 {[A-Z]+}
    set re3 {[1-9]+}
    if {[regexp $re1 $pw] && [regexp $re2 $pw] && [regexp $re3 $pw]} {
        return $pw
    } else {
        return [genPw $pwlen]
    }
}
# mouse tracker
proc addCoords {x y} {
    append ::track $x $y
}
proc newPw {} {
    set ::pswd [genPw]
    mkKey
    .mw.e1 icursor end
}

global pswd key botikkey tcl_platform track
set track {}; # mouse track

initValues
catch {destroy .mw}

catch {
   if { $tcl_platform(platform) == "windows" } {
      set icon img/botik_kg.ico
   } else {
      set icon [lindex $botikkey(YELLOW) 0 ]
   }
   wm  iconbitmap . $icon
   wm  title      . "     "
}

labelframe .mw -text $botikkey(version) -font {Helvetica 10 bold}
pack  .mw
focus -force .mw
label .mw.t1 -font {Helvetica 10} -justify left -text \
                   "    $botikkey(version).\
                   \n\
                   \n        \
                   \n(       ).\
                   \n\
                   \n     :\
                   \n\
                   \n1. ,       ,\
                   \n         \
                   \n ."

frame  .mw.f
label  .mw.f.l -font {Helvetica 10} -justify left -text "\n2.      :"
button .mw.f.rpw -text " " -command {newPw} -takefocus 0

set pswd [genPw]
mkKey1

entry .mw.e1 -validate all -vcmd mkKey -textvariable pswd -state normal -font {Helvetica 11}

label .mw.t2 -font {Helvetica 10} -justify left -text \
                  "\n3.   ,     \
                  \n(      ):"

entry .mw.e2 -textvariable key -state readonly -font {Helvetica 11}

label .mw.t3 -font {Helvetica 10} -justify left -text \
                  "\n4.         ( )\
                  \n     .    \
                  \n  (, CD-, FLASH/USB-  ..),  \
                  \n        ,  \
                  \n    ( ,     )."

frame .mw.bs

pack .mw.t1 -side top -anchor w -padx 5
pack .mw.f -side top -anchor w
pack .mw.f.l -anchor w -side left -padx 5
pack .mw.f.rpw -side left -padx 20 -anchor s -pady 3
pack .mw.e1 .mw.t2 .mw.e2 .mw.t3 .mw.bs -side top -anchor w -padx 5 -fill x

setImgButtons .mw.bs 2 2 left w [subst { 
      close $myImages(noNormal)    {exit}
      save  $myImages(fnameNormal) writeKey
} ]

bind . <Motion> "addCoords %x %y"
focus -force .mw.e1
update
