#!/usr/bin/wish

# This is the client side impl. of speed testing framework.
# Follow this link for detailed description:
#  http://wiki.botik.ru/Nadmin/BotikSpeed 
# Version 0.1 released 2006-12-18.
# Version 1.0 released 2011-02-01.
# Version 1.1 released 2011-06-01.

# Copyright (c) 2006-2014, Botik Technologies, Ltd.
# See license.txt for more information.

namespace eval speed {
variable ddir ""
proc showHelp {} {
    set text " $speed::pname    (In/Out)     ,\
        (  ).\
       \
       ( /).\
         ,\
    ,      .\
  ,     \
   ,     \"$::net(name)\".\
      $speed::timeo .\
\n               .\
       (.jpg)\
    (.txt)    \"  \".\
      \
  ; ,  http://img.pereslavl.ru"
    set w .about
    catch {destroy $w}
    toplevel $w -class Dialog
    wm title $w "  $speed::title"
    wm iconname $w Dialog
    if {[winfo viewable [winfo toplevel [winfo parent $w]]] } {
        wm transient $w [winfo toplevel [winfo parent $w]]
    }
    set windowingsystem [tk windowingsystem]
    if {$windowingsystem eq "aqua"} {
        ::tk::unsupported::MacWindowStyle style $w moveableModal {}
    }
    frame $w.bot
    frame $w.top
    if {$windowingsystem eq "x11"} {
        $w.bot configure -relief raised -bd 1
        $w.top configure -relief raised -bd 1
    }
    pack $w.bot -side bottom -fill both
    pack $w.top -side top -fill both -expand 1
    grid anchor $w.bot center
    option add *Dialog.msg.wrapLength 5i widgetDefault
    #option add *Dialog.msg.font TkCaptionFont widgetDefault
    label $w.msg -justify left -text $text -font {Times 11}
    pack $w.msg -in $w.top -side right -expand 1 -fill both -padx 3m -pady 3m
    set i 0
    set but "OK"
    set default 0
    button $w.button$i -text $but -command "destroy $w" -font {Times 11}
    if {$i == $default} {
        $w.button$i configure -default active
    } else {
        $w.button$i configure -default normal
    }
    grid $w.button$i -in $w.bot -column $i -row 0 -sticky ew \
        -padx 10 -pady 4
    grid columnconfigure $w.bot $i
    # We boost the size of some Mac buttons for l&f
    if {$windowingsystem eq "aqua"} {
        set tmp [string tolower $but]
        if {$tmp eq "ok" || $tmp eq "cancel"} {
            grid columnconfigure $w.bot $i -minsize 90
        }
        grid configure $w.button$i -pady 7
    }
    incr i
    if {$default >= 0} {
        bind $w <Return> "destroy $w"
    }
    wm withdraw $w
    update idletasks
    set x [expr {[winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
        - [winfo vrootx [winfo parent $w]]}]
    set y [expr {[winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
        - [winfo vrooty [winfo parent $w]]}]
    if {$x < 0} {set x 0}
    if {$y < 0} {set y 0}
    wm maxsize $w [winfo screenwidth $w] [winfo screenheight $w]
    wm geometry $w +$x+$y
    wm deiconify $w
}
proc speedFmt {x} {
    set k 0
    set y $x
    if {$y < 1024} {
        return [format "%.0f /" $y]
    }
    set labels [list " " "/" "/" "/" "/"]
    for {set k 0} { ($k <= [llength $labels]) && (double($y) > 999.0) } {incr k} {
        set y [expr {double($y) / 1024.0}]
    }
    return [format "%.2f %s" $y [lindex $labels $k]]
}
proc windowsPlatform {} {
   return [expr {$::tcl_platform(platform) eq "windows"}]
}
proc unixPlatform {} {
   return [expr {$::tcl_platform(platform) eq "unix" && $::tcl_platform(os) ne "Darwin"}]
}
proc alertError {msg} {
    tk_messageBox -icon error -type ok -title "" -message "$msg"
}
proc alert {msg} {
    tk_messageBox -icon info -type ok -title "" -message "$msg"
}
proc mbit_fmt {bytes} {
    return [format "%.2f /" [expr {($bytes * 8) / 1000000.0}]]
}
proc kbit_fmt {bytes} {
    return [format "%.2f /" [expr {($bytes * 8) / 1000.0}]]
}
proc kbits {bytes} {
    return [format "%.2f" [expr {($bytes * 8) / 1000.0}]]
}
proc gettime {} {
    if {[unixPlatform]} {
        return [format "%.3f" [exec date +%s.%N]]
    }
    return [format "%.3f" [expr {[clock clicks -milliseconds]/1000.0}]]
}
proc debug {msg} {
    #if {$speed::debug} {catch {puts $msg}}
    if {$speed::debug} {
        catch {
            set fd [open speeddbg.txt a]
            puts $fd "\[[gettime]] $msg"
            close $fd
        }
    }
}
# time1 - client's send time
# time2 - server's recv time
# time3 - server's send time
# time4 - client's recv time
# correct time to be set on client =
#   $time3 + (($time4 - $time1) - ($time3 - $time2)) / 2.0
proc writeData {chan step} {
    variable randstr
    variable sendsize
    variable time1
    variable btime
    variable timeo
    variable size_u
    variable size_d
    variable time_u
    variable time_d
    variable len0
    catch {fileevent $chan writable {}}
    if {$step > 0} {
        set n [expr {int(pow(2,$step))}]
        set data [string repeat $randstr $n]
        set sendsize [expr {$len0*$n}]
    } else {
        set data "sync"
        set sendsize 4
    }
    if {[gettime] - $btime >= $timeo/2.0 && $size_u + $size_d != 0 && $time_u + $time_d != 0} {
        set curspeed [expr {($size_u+$size_d)/($time_u+$time_d)}]; # current up+down speed
        set size [expr {$sendsize * 2}]; # send+recv size
        set needtime [expr {double($size)/$curspeed}]; # time needed to complete transaction
        set remtime [expr {$timeo - ([gettime] - $btime)}]; # remaining time
        debug "remain: $remtime sec., need: $needtime sec."
        if {$needtime > 0 && $needtime - $remtime >= 2.0} {
            stopSpeedTest; # do not send data if we don't have time to recv it
            return
        }
    }
    set time1 [gettime]
    if {[catch {
        puts $chan $data; flush $chan
    } code]} {
        debug $code
        stopSpeedTest
        return
    }
    debug "sent data (step=$step, size=$sendsize)"
    catch {fileevent $chan readable [list speed::readData $chan $step]}
}
proc readData {chan step} {
    variable size_u
    variable size_d
    variable time_u
    variable time_d
    variable time1
    variable timeshift
    variable stepdone$step
    variable sendsize
    variable stepsdone
    if {[eof $chan] || [catch {gets $chan data}]} {
        catch {close $chan}
        debug "connection closed"
        set stepdone$step 1
        return
    }
    if {![llength $data]} {return}
    set time4 [gettime]
    catch {fileevent $chan readable {}}
    set time2 ""
    set time3 ""
    set recvsize [string length $data]
    debug "recv data (step=$step, size=$recvsize)"
    foreach {dat_ time2 time3} $data {break}
    if {[string is double -strict $time2] && [string is double -strict $time3]} {
        if {$step == 0} {
            set truetime [expr {$time3 + (($time4 - $time1) - ($time3 - $time2)) / 2.0}]; # SNTP-like approach
            set timeshift [expr {[gettime] - $truetime}]
            debug "truetime=$truetime,mytime=[gettime]\ntimeshift=$timeshift"
        } else {
            debug "time1=[expr $time1 - $timeshift]\ntime2=$time2\ntime3=$time3\ntime4=[expr $time4 - $timeshift]"
            set upltime  [expr {$time2 - ($time1 - $timeshift)}]; # time spent to upload the data to server
            set downtime [expr {$time4 - $timeshift - $time3}]; # time spent to download the data from server
            if {$upltime < 0} {set upltime 0}
            if {$downtime < 0} {set downtime 0}
            set size_u [expr {$size_u + $sendsize}]
            set size_d [expr {$size_d + $recvsize}]
            set time_u [expr {$time_u + $upltime}]
            set time_d [expr {$time_d + $downtime}]
        }
        incr stepsdone
    }
    set stepdone$step 1
}
proc random_string {length} {
    set chars "abcdefghijklmnopqrstuvwxyz"
    set range [expr {[string length $chars]-1}]
    set txt ""
    for {set i 0} {$i < $length} {incr i} {
       set pos [expr {int(rand()*$range)}]
       append txt [string index $chars $pos]
    }
    return $txt
}
proc getSpeedResult {} {
    variable size_u
    variable size_d
    variable time_u
    variable time_d
    variable len0
    variable stepsdone
    if {$stepsdone < 2} {return}; # at least 2 send/recv transactions must be completed
    set res {}
    if {$size_u > $len0} {
        if {$time_u > 0} {
            set speed [expr {$size_u / $time_u}]
        } else {
            set speed >$size_u
        }
        lappend res [list upload $speed]
    }
    if {$size_d > $len0} {
        if {$time_d > 0} {
            set speed [expr {$size_d / $time_d}]
        } else {
            set speed >$size_d
        }
        lappend res [list download $speed]
    }
    debug "time_u=$time_u\ntime_d=$time_d\nsize_u=$size_u\nsize_d=$size_d"
    return $res
}
proc convert_result {res} {
    set str ""
    foreach elt $res {
        foreach {type spd} $elt {break}
        set less ""
        if {[string index $spd 0] eq "<"} {
            set less " "
            set spd [string range $spd 1 end]
        } elseif {[string index $spd 0] eq ">"} {
            set less " "
            set spd [string range $spd 1 end]
        }
        if {$type eq "upload"} {
            append str "   :      $less[mbit_fmt $spd]\n"
        }
        if {$type eq "download"} {
            append str "   :    $less[mbit_fmt $spd]\n"
        }
    }
    return $str
}
proc convert_result2 {res} {
    set str ""
    foreach elt $res {
        foreach {type spd} $elt {break}
        set less ""
        if {[string index $spd 0] eq "<"} {
            set less " "
            set spd [string range $spd 1 end]
        } elseif {[string index $spd 0] eq ">"} {
            set less " "
            set spd [string range $spd 1 end]
        }
        if {$type eq "upload"} {
            append str "   . : $less[mbit_fmt $spd]\n"
        }
        if {$type eq "download"} {
            append str "   . :  $less[mbit_fmt $spd]\n"
        }
    }
    return $str
}
proc stopSpeedTest {} {
    variable speedaid
    variable chan
    variable maxsteps
    variable stopped
    if {$stopped} {return}
    set stopped 1
    catch {after cancel $speedaid}
    catch {close $chan}
    debug "stopped"
    set step 0
    while {$step <= $maxsteps} {
        variable stepdone$step
        set stepdone$step 1
        incr step
    }
}
proc do_speed_test {{srvnum 1}} {
    initVars
    variable randstr
    variable speedaid
    variable chan
    variable size_u
    variable size_d
    variable time_u
    variable time_d
    variable timeo
    variable timeshift
    variable maxsteps
    variable server1
    variable server2
    variable port
    variable btime
    variable len0
    for {set i 0} {$i <= $maxsteps} {incr i} {
        variable stepdone$i
    }
    if {![string is integer -strict $srvnum] || $srvnum < 1 || $srvnum > 2} {return}
    set randstr [random_string $len0]
    set btime [gettime]
    # set timer that expires in 20 seconds and eventually stops speedtesting
    set speedaid [after [expr {$timeo*1000}] speed::stopSpeedTest]
    if {[catch {
        set chan [socket -async [set server$srvnum] $port]
    } code]} {
        debug $code
        stopSpeedTest; return
    }
    fconfigure $chan -buffering line -blocking 0
    set step 0
    # step=0: determine time shift (skew) between client and server clocks
    # step>0: send and recv data, obtain timings
    while {$step <= $maxsteps} {
        if {[catch {
            fileevent $chan writable [list speed::writeData $speed::chan $step]
        } code]} {
            debug $code
            break
        }
        vwait speed::stepdone$step
        if {$step == 0 && $timeshift eq ""} {
            stopSpeedTest; return
        }
        incr step
    }
    stopSpeedTest
    return [getSpeedResult]
}
proc initVars {} {
    variable version 1.1
    variable pname   "BotikSpeed"
    variable title   "$pname v$version"
    variable server1 speedtest.botik.ru; # Moscow node
    variable server2 speedtest2.botik.ru; # local node
    variable port    21564
    variable ip1     62.140.250.221; # default IP (BotikTester needs it)
    variable presult 0; # parsable result
    variable gresult  ; # label
    variable randstr   ""
    variable speedaid  ""
    variable chan      ""
    variable timeshift ""
    variable timeo     20; # speed test timeout (in seconds)
    variable sendsize  0
    variable size_u    0
    variable size_d    0
    variable time_u    0
    variable time_d    0
    variable time1     0
    variable maxsteps  8; # max. num. of steps
    for {set i 0} {$i <= $maxsteps} {incr i} {
        variable stepdone$i 0
    }
    variable tspeed
    variable uname
    variable hname
    variable remain
    variable btime; # speed test begin time
    variable fintime 0
    variable stopped 0
    variable len0 [expr {32*1024}]
    variable debug 0
    variable stepsdone 0
    if {![info exists ::ping(name)]} {
        set ::botiktools(progname) $pname
    }
}
proc tariff_speed_fmt {ts} {
    variable tspeed
    set res ""
    if {[llength $ts] && $ts > 0} {
        set tspeed $ts
        set fmt [format "%.2f /" [expr {$ts / 1000.0}]]
        append res "  : $fmt\n"
    } elseif {[info exists tspeed] && $tspeed > 0} {
        set fmt [format "%.2f /" [expr {$tspeed / 1000.0}]]
        append res "  : $fmt\n"
    }
    return $res
}
proc get_tariff_speed {} {
    variable uname
    variable hname
    set data ""
    catch {set data [informer_data]}
    set text [encoding convertfrom koi8-r $data]
    set spl  [split $text \n]
    set res  0
    set par  ""
    set val  ""
    foreach line $spl {
        if {![llength $line]} {continue}
        set pv [split $line :]
        foreach {- par val} $pv {break}
        if {$par eq "TARIFF_SPEED"} {set res $val}
        if {$par eq "UserName"} {set uname $val}
        if {$par eq "HostName"} {set hname $val}
    }
    return $res
}
proc updateStatus {} {
    variable remain
    variable timeo
    variable presult
    variable gresult
    if {$presult != 0 || $remain < 0} {return}
    if {$remain == 0} {
        set gresult "...\n"
    } elseif {$remain >= 10} {
        set gresult "...\n $remain ."
    } else {
        set gresult "...\n $remain . "
    }
    if {$remain < $timeo} {update idletasks}
    incr remain -1
    after 1000 speed::updateStatus
}
proc outer_speed {} {
    variable gresult
    variable presult
    variable fintime
    variable remain
    variable timeo
    enable_buttons 0
    update_canvas {}
    set presult 0
    set remain $timeo
    updateStatus; update
    set presult [do_speed_test]
    set fintime [clock seconds]
    if {![llength $presult] || $presult==0} {
        set gresult ".\n"
    } else {
        set gresult "\n"
    }
    update
    update_canvas [getresult]
    enable_buttons 1
    update
}
proc getresult {} {
    variable presult
    variable uname
    variable hname
    variable fintime
    if {![llength $presult] || $presult==0} {return}
    set uinfo ""
    set ts [get_tariff_speed]
    if {[info exists uname] && [llength $uname]} {
        set uinfo ": $uname\n: $hname\n"
    }
    set res $uinfo[convert_result $presult]
    append res [tariff_speed_fmt $ts]
    if {$fintime == 0} {set fintime [clock seconds]}
    set date [clockFormat ": %d.%m.%Y %H:%M:%S\n" $fintime]
    append res $date
    return $res
}
proc getTextDim {text font} {
    set c .c
    set tag tmptxt
    $c create text -100 -100 -text $text -font $font -tags $tag
    set bb [$c bbox $tag]
    $c delete $tag
    foreach {x1 y1 x2 y2} $bb {break}
    set tw [expr {int($x2-$x1)}]
    set th [expr {int($y2-$y1)}]
    return [list $tw $th]
}
proc textB1Press {w x y} {
    $w select clear
    $w icursor speedtext @$x,$y
    $w focus speedtext
    focus $w
    $w select from speedtext @$x,$y
}
proc textB1Move {w x y} {
    $w select to speedtext @$x,$y
    textCopy $w
}
proc textCopy {w} {
    set sel ""
    catch {set sel [selection get -displayof $w]}
    if {[llength $sel]} {
        clipboard clear
        clipboard append $sel
    }
}
proc textSelectAll {w} {
    $w icursor speedtext 0
    $w select from speedtext 0
    $w select to speedtext end
    textCopy $w
}
proc update_canvas {txt} {
    set c .c
    $c delete all
    if {![llength $txt]} {
        $c configure -width 1 -height 1
        return
    }
    set font {Arial 12 bold}
    set dims [getTextDim $txt $font]
    foreach {tw th} $dims {break}
    set w [expr {$tw + 15}]
    set h [expr {$th + 15}]
    $c configure -width $w -height $h
    $c create text 10 10 -text $txt -font $font -anchor nw -tags speedtext
    $c bind speedtext <1> "speed::textB1Press $c %x %y"
    $c bind speedtext <B1-Motion> "speed::textB1Move $c %x %y"
    $c bind speedtext <Shift-1> "$c select adjust current @%x,%y; speed::textCopy $c"
    $c bind speedtext <Shift-B1-Motion> "speed::textB1Move $c %x %y"
    $c bind speedtext <Control-c> "speed::textCopy $c"
    $c bind speedtext <Control-Insert> "speed::textCopy $c"
    $c bind speedtext <Control-a> "speed::textSelectAll $c"
    $c bind speedtext <Double-1>  "speed::textSelectAll $c"
}
proc getFileName {} {
    variable uname
    variable fintime
    if {$fintime == 0} {set fintime [clock seconds]}
    set time [clockFormat %d.%m.%Y_%H-%M-%S $fintime]
    set res "speedtest"
    if {[info exists uname] && [llength $uname]} {append res _$uname}
    append res _$time
    return $res
}
proc set_default_dir {dir} {
    variable ddir
    if {[file isdirectory $dir]} {
        set ddir $dir
    }
}
# see proc desktop_dir in botikUts.tcl
proc default_dir {} {
    variable ddir
    if {$ddir ne ""} {return $ddir}
    set dir ~
    if {[windowsPlatform]} {
        catch {
            package require twapi
            set dir [twapi::get_shell_folder desktopdirectory]
        }
    }
    set ddir $dir
    return $dir
}
proc takeScreenshot {c} {
    variable size_u
    variable size_d
    variable time_u
    variable time_d
    variable presult
    if {![llength $presult] || $presult == 0} {return}
    if {[catch {
        package require Img
    }]} {return}
    set img  "canvasPhoto"
    set name [getFileName]
    set types {
        {"JPEG file" .jpg JPEG}
        {"Text file" .txt TEXT}
    }
    set fileName [tk_getSaveFile -filetypes $types \
        -parent $c -initialdir [default_dir] \
        -initialfile $name -defaultextension .jpg]
    if {![llength $fileName]} {return}
    set_default_dir [file dirname $fileName]
    update
    set ext [file extension $fileName]
    if {$ext eq ".jpg"} {
        image create photo $img -format window -data $c
        $img write $fileName -format JPEG
        image delete $img
        $c delete $img
        $c dtag $img
        return
    }
    if {$ext ne ".txt"} {append fileName ".txt"}
    set res [getresult]
    if {[llength $res]} {
        append res ": $size_u .\n"
        append res ": $size_d .\n"
        append res " : [format %.3f $time_u] .\n"
        append res " : [format %.3f $time_d] .\n"
    }
    set fd [open $fileName w]
    puts $fd $res
    close $fd
}
proc enable_buttons {enable} {
    catch {
    if {$enable} {
        .speed configure -state normal
        .screen configure -state normal
    } else {
        .speed configure -state disabled
        .screen configure -state disabled
    }
    }
}
proc showGUI {} {
    variable gresult
    wm title . $speed::title
    catch {wm iconbitmap . -default [file join img botikspeed.ico]}
    set font {Helvetica 11}
    label .name -text " " -font {Helvetica 14 bold} -foreground blue -anchor n
    label .howto -text ",   \n    \"$::net(name)\":" -font {Helvetica 10}
    button .speed -text " " -width 18 -font $font -command speed::outer_speed
    button .screen -text "  " -width 15 -font $font -command "speed::takeScreenshot .c" -state disabled
    set gresult "\n"
    label .gresult -textvariable speed::gresult -font $font
    canvas .c -width 1 -height 1
    frame .buttons
    button .buttons.help -text " " -command speed::showHelp
    button .buttons.exit -text "" -command exit
    pack .name -padx 10 -pady 5 -fill x -expand 1 -side top
    pack .howto .speed -padx 10 -pady 5 -side top
    pack .gresult .c .screen .buttons -padx 5 -pady 2 -side top
    pack .buttons.help .buttons.exit -side left -pady 5
    focus -force .
    bind . <KeyPress-Return> ".speed invoke"
    wm resizable . 0 0
}
}; # end of speed:: namespace

if {![info exists ::ping(name)]} {
    # if run as standalone application
    source net.tcl
    if {[speed::windowsPlatform]} {
        catch { package require twapi }
        catch { package require informer }
        package require tile
        namespace import -force ttk::*
        catch { informer_start "https://nadmin.$::net(domain)/~nadmin/whoami.cgi" 60 }
    }
    speed::initVars
    speed::showGUI
} else {speed::initVars}
